use std::path::{self, Path, PathBuf};

use anyhow::{ensure, Result};
use clap::Subcommand;

use crate::state::write_projects_file;
use crate::{dirs, Project, Projects};


#[derive(Debug, Clone, clap::ValueEnum)]
pub enum ShellType {
    Nu,
}

impl ShellType {
    pub fn integration(&self) -> String {
        let template = match self {
            Self::Nu => include_str!("./integrations/nushell.nu"),
        };

        let file_path = dirs::selected_project_file().to_str().unwrap();
        let current_exe = std::env::current_exe().unwrap();
        let current_exe = current_exe.to_str().unwrap();

        template
            .replace("{%SELECTED_PROJECT_FILE%}", file_path)
            .replace("{%PROJ_EXE%}", current_exe)
    }
}


#[derive(Debug, Subcommand)]
pub enum Command {
    /// Add a project
    Add { path: PathBuf },

    /// Remove a project
    Remove { path: PathBuf },

    /// List existing projects
    List,

    /// Initialize shell integration
    Init { shell: ShellType },
}

impl Command {
    pub fn execute(&self, mut projects: Projects) -> Result<()> {
        match self {
            Command::Add { path } => {
                add_project(&mut projects, path)?;
            }
            Command::Remove { path } => {
                remove_project(&mut projects, path)?;
            }
            Command::List => {
                list_projects(&projects)?;
            }
            Command::Init { shell } => {
                print_shell_integration(shell);
            }
        }

        Ok(())
    }
}


fn print_shell_integration(shell: &ShellType) {
    println!("{}", shell.integration());
}


fn add_project<P>(projects: &mut Projects, path: P) -> Result<()>
where
    P: AsRef<Path>,
{
    let path = path::absolute(path)?;
    ensure!(path.is_dir(), "Project path does not exists");

    let project = Project::from(path);

    ensure!(
        !projects.list.contains(&project),
        "Project path already registered"
    );

    projects.list.push(project);
    write_projects_file(projects)?;
    println!("Added {}", projects.list.last().unwrap().path().display());

    Ok(())
}


fn remove_project<P>(projects: &mut Projects, path: P) -> Result<()>
where
    P: AsRef<Path>,
{
    let path = path::absolute(path)?;
    let project = Project::from(path);

    ensure!(
        projects.list.contains(&project),
        "Project path not in registry"
    );

    let idx =
        projects.list.iter().enumerate().find_map(
            |(idx, elem)| {
                if elem == &project {
                    Some(idx)
                } else {
                    None
                }
            },
        );

    if let Some(idx) = idx {
        let proj = projects.list.remove(idx);
        write_projects_file(projects)?;
        println!("Removed {}", proj.path().display());
    }

    Ok(())
}


fn list_projects(projects: &Projects) -> Result<()> {
    for (idx, project) in projects.list.iter().enumerate() {
        println!("{}: {} ({})", idx + 1, project.name(), project.path().display())
    }

    Ok(())
}
