use std::collections::VecDeque;
use std::sync::mpsc;

use anyhow::Result;
use crossterm::event::{self, Event, KeyCode, KeyEvent, KeyModifiers};
use ratatui::layout::{Constraint, Layout, Rect};
use ratatui::style::{Color, Style};
use ratatui::text::Span;
use ratatui::widgets::{Block, Borders, Cell, Paragraph, Row, Table, TableState};
use ratatui::Frame;
use tracing::trace;
use tui_input::{Input, InputRequest};

use crate::state::{clear_selected_project_file, write_selected_project_file};
use crate::{Project, Projects};


#[derive(Debug, Clone, PartialEq, Eq)]
enum Message {
    Exit,
    Noop,
    SelectPrevious,
    SelectNext,
    Confirm,
    SearchUpdate,
}


#[derive(Debug, Default)]
enum Mode {
    #[default]
    Search,
    Rename,
}


struct State {
    projects: Projects,
    search: Input,
    mode: Mode,
    should_exit: bool,
    project_table: TableState,
    filtered_projects: Vec<Project>,
    selected_project: Option<Project>,
}

impl State {
    pub fn new(projects: Projects) -> Self {
        let project_table = TableState::new().with_selected(Some(0));

        Self {
            projects,
            search: Input::default(),
            mode: Mode::default(),
            should_exit: false,
            project_table,
            filtered_projects: Vec::new(),
            selected_project: None,
        }
    }
}


#[tracing::instrument(skip_all)]
pub fn run(projects: Projects) -> Result<()> {
    clear_selected_project_file()?;

    let mut term = ratatui::init();
    let mut state = State::new(projects);

    let (mut msg_tx, mut msg_rx) = mpsc::channel();

    loop {
        handle_messages(&mut state, &mut msg_rx)?;

        if state.should_exit {
            break;
        }

        term.draw(|frame| draw(&mut state, frame))?;

        if let Event::Key(key_event) = event::read()? {
            handle_key_event(&mut state, &mut msg_tx, key_event);
        }
    }

    ratatui::restore();

    if let Some(selected_project) = state.selected_project {
        // hacky stderr abuse
        eprintln!(
            "{}-cd:{}",
            env!("CARGO_BIN_NAME"),
            selected_project.path().display()
        );
        write_selected_project_file(&selected_project)?
    }

    Ok(())
}


fn handle_key_event(state: &mut State, tx: &mut mpsc::Sender<Message>, event: KeyEvent) {
    let msg = match (&state.mode, event.modifiers, event.code) {
        (Mode::Search, KeyModifiers::CONTROL, KeyCode::Char('p' | 'k') | KeyCode::Up) => {
            Message::SelectPrevious
        }
        (Mode::Search, KeyModifiers::CONTROL, KeyCode::Char('n' | 'j') | KeyCode::Down) => {
            Message::SelectNext
        }
        (Mode::Search, KeyModifiers::NONE, KeyCode::Esc) => Message::Exit,
        (Mode::Search, KeyModifiers::NONE | KeyModifiers::SHIFT, KeyCode::Char(c)) => {
            state.search.handle(InputRequest::InsertChar(c));
            Message::SearchUpdate
        }
        (Mode::Search, KeyModifiers::NONE, KeyCode::Backspace) => {
            state.search.handle(InputRequest::DeletePrevChar);
            Message::SearchUpdate
        }
        (_, _, KeyCode::Enter) => Message::Confirm,

        _ => Message::Noop,
    };

    tx.send(msg).unwrap();
}


fn handle_messages(state: &mut State, rx: &mut mpsc::Receiver<Message>) -> Result<()> {
    let msg = match rx.try_recv() {
        Ok(msg) => msg,
        Err(mpsc::TryRecvError::Empty) => {
            return Ok(());
        }
        Err(mpsc::TryRecvError::Disconnected) => {
            state.should_exit = true;
            return Ok(());
        }
    };

    match msg {
        Message::Exit => {
            state.should_exit = true;
            state.selected_project = None;
        }
        Message::SelectNext => {
            state.project_table.select_next();
        }
        Message::SelectPrevious => {
            state.project_table.select_previous();
        }
        Message::Confirm => {
            if let Some(selected) = state.project_table.selected() {
                if let Some(project_path) = state.filtered_projects.get(selected) {
                    state.should_exit = true;
                    state.selected_project = Some(project_path.clone());
                };
            }
        }
        Message::SearchUpdate => {
            state.project_table.select_first();
        }
        _ => (),
    }

    Ok(())
}


fn draw(state: &mut State, frame: &mut Frame) {
    let block = Block::default().borders(Borders::ALL);
    frame.render_widget(&block, frame.area());

    match state.mode {
        Mode::Search => {
            let layout = Layout::vertical([Constraint::Fill(1), Constraint::Max(2)]);
            let inner_area = block.inner(frame.area());
            let layout_rects = layout.split(inner_area);
            draw_list(state, frame, layout_rects[0]);
            draw_search(state, frame, layout_rects[1]);
        }
        Mode::Rename => {
            let layout = Layout::vertical([Constraint::Fill(1)]);
            let inner_area = block.inner(frame.area());
            let layout_rects = layout.split(inner_area);
            draw_list(state, frame, layout_rects[0]);
        }
    }
}


fn draw_search(state: &mut State, frame: &mut Frame, area: Rect) {
    const PROMPT_PREFIX: &str = "Search: ";

    let search_input = &state.search;

    let scroll_offset = (0, search_input.visual_scroll(area.width as usize) as u16);

    let prompt = format!("{PROMPT_PREFIX}{}", search_input.value());
    let search = Paragraph::new(prompt).scroll(scroll_offset);
    frame.render_widget(search, area);

    let cursor_pos = (
        area.x + (search_input.visual_cursor() + PROMPT_PREFIX.len()) as u16,
        area.y,
    );
    frame.set_cursor_position(cursor_pos)
}


fn draw_list(state: &mut State, frame: &mut Frame, area: Rect) {
    let search_value = state.search.value();
    let projects = &state.projects.list;

    state.filtered_projects = if search_value.is_empty() {
        projects.clone()
    } else {
        let mut filtered = VecDeque::new();

        for project in projects.iter() {
            match fuzzy_search(search_value, &project.name()) {
                Some(indices) => {
                    trace!(?search_value, ?project, ?indices);
                    filtered.push_front(project.clone());
                }
                None => {
                    let path = project.path().to_str().unwrap();
                    if let Some(indices) = fuzzy_search(search_value, path) {
                        trace!(?search_value, ?project, ?indices);
                        filtered.push_back(project.clone());
                    }
                }
            }
        }

        Vec::from(filtered)
    };

    let rows = state.filtered_projects.iter().map(|project| {
        let name = project.name();
        let path_style = Style::new().fg(Color::DarkGray);
        let path_span = Span::styled(project.path().display().to_string(), path_style);
        let name_span = Span::from(name);

        Row::new([Cell::new(name_span), Cell::new(path_span)])
    });

    let widths = [Constraint::Min(20), Constraint::Fill(1)];

    let table_highlight = Style::default().fg(Color::Blue);

    let table = Table::new(rows, widths).row_highlight_style(table_highlight);

    frame.render_stateful_widget(table, area, &mut state.project_table);
}


fn fuzzy_search(search: &str, text: &str) -> Option<Vec<usize>> {
    let mut found_indices = Vec::new();
    let mut start_idx = 0;

    let mut case_sensitive = false;

    for (ch_idx, ch) in search.char_indices() {
        if ch_idx == 0 {
            case_sensitive = ch.is_uppercase();
        }

        let remaining = &text[start_idx..];

        let mut found = false;

        for (byte_idx, txt_ch) in remaining.char_indices() {
            let matching = if case_sensitive {
                ch == txt_ch
            } else {
                ch.to_lowercase().cmp(txt_ch.to_lowercase()).is_eq()
            };

            if matching {
                found_indices.push(start_idx + byte_idx);
                start_idx += byte_idx + txt_ch.len_utf8();
                found = true;
                break;
            }
        }

        if !found {
            return None;
        }
    }

    if search.is_empty() {
        None
    } else {
        Some(found_indices)
    }
}


#[cfg(test)]
mod test {
    use super::*;

    #[test]
    fn fuzzy_test() {
        assert_eq!(fuzzy_search("st", "st"), Some(vec![0, 1]));
        assert_eq!(fuzzy_search("st", "some text"), Some(vec![0, 5]));
        assert_eq!(fuzzy_search("st", "nothing"), None);
        assert_eq!(fuzzy_search("st", ""), None);
        assert_eq!(fuzzy_search("", "nonempty"), None);
    }
}
